/*
* Climate box control for beer brewing.
*
* The climate box utilizes a peltier element for cooling and a terrarium heater for
* heating.  The cooler has a fan attached that is turned off later than the cooler
* itself is turned off.  Also, the cooler and the heater remain in "on" or "off"
* states for at least a few minutes whenever the state toggles in order to reduce
* wear.
*
* An eight character LCD displays the current temperature, the target temperature,
* and a glyph indicating whether the heater or the cooler is on.
*
* The temperature is measured by sampling an MCP9700, and the target temperature is
* set using two pushbuttons.
*
* The current temperature, the target temperature, and the heater/cooler activity
* are periodically written to serial port 1 at 115200 bps 8N1.  The climate box can
* be disabled or enabled by writing '0' or '1' to its serial input.
*
*
* This program is free software: you can redistribute it and/or modify it under
* the terms of the GNU General Public License as published by the Free Software
* Foundation, either version 3 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful, but WITHOUT ANY
* WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
* PARTICULAR PURPOSE.  See the GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License along with this
* program.  If not, see <http://www.gnu.org/licenses/>.
*/


/* Libraries. */
#include <LiquidCrystal.h>
#include <EEPROM.h>
#include <PID_v1.h>
#include <avr/wdt.h>


#define DEBUG
//#define DEBUG_MEASURE_ONLY


/* Initialize the LCD library with the following the interface pins:
  * LCD RS pin to digital pin 12
  * LCD Enable pin to digital pin 11
  * LCD D4 pin to digital pin 5
  * LCD D5 pin to digital pin 4
  * LCD D6 pin to digital pin 3
  * LCD D7 pin to digital pin 2
*/
LiquidCrystal lcd_display( 12, 11, 5, 4, 3, 2 );

typedef enum custom_character {
  ARROW_DOWN = 0,
  ARROW_UP   = 1,
  CHECKMARK  = 2
};
byte arrow_down[ 8 ] = {
  B00100,
  B00100,
  B00100,
  B00100,
  B10101,
  B01110,
  B00100,
};
byte arrow_up[ 8 ] = {
  B00100,
  B01110,
  B10101,
  B00100,
  B00100,
  B00100,
  B00100,
};
byte checkmark[ 8 ] = {
  B00000,
  B00000,
  B00001,
  B00010,
  B10100,
  B01000,
  B00000,
};

const int display_update_interval = 5000;

const int cooler_pin           = 6;
const int cooler_fan_pin       = 7;
const int heater_pin           = 8;
const int temperature_down_pin = 9;
const int temperature_up_pin   = 10;
const int temperature_pin      = 0;
const int led_pin              = 13;

float       target_temperature  = 15.0;
const float minimum_temperature = 1.0;
const float maximum_temperature = 30.0;
const float default_temperature = 15.0;
const float temperature_lower_window = 0.2;
const float temperature_lower_window_low = 0.7;
const float temperature_upper_window = 0.1;
const float temperature_upper_window_high = 0.7;

/* Cooler controller. */
bool         cooling          = false;
const long   cool_period      = 30000l;
const long   lower_cool_period = 8000l;
const long   upper_cool_period = cool_period - 5000l;
double       cooler_setpoint;
double       cooler_input;
double       cooler_output;
const double cooler_kp        = 1000.0;
const double cooler_ki        = 50.0;
const double cooler_kd        = 25.0;
PID cooler_PID( &cooler_input, &cooler_output, &cooler_setpoint, cooler_kp, cooler_ki, cooler_kd, REVERSE );

/* Heater controller. */
bool         heating          = false;
const long   heat_period      = 3000l;
const long   lower_heat_period = 20l;
const long   upper_heat_period = heat_period - 20l;
double       heater_setpoint;
double       heater_input;
double       heater_output;
const double heater_kp        = 25.0;
const double heater_ki        = 3.0;
const double heater_kd        = 1.0;
PID heater_PID( &heater_input, &heater_output, &heater_setpoint, heater_kp, heater_ki, heater_kd, DIRECT );

float current_temperature;

static bool cooler_enabled = true;
static bool heater_enabled = true;
const float setpoint_window = 0.3;


/* Cooler fan control. */
bool       cooldown_fan         = false;
long       cooler_fan_timestamp = 0l;
const long cooler_fan_delay     = 30000l;

const long serial_baud_rate       = 115200;
#ifdef DEBUG
const long status_update_interval = 10*60000l;
#else
const long status_update_interval = 30000l;
#endif

#ifdef DEBUG
int mcp9700_voltage;
#endif

typedef enum key_press {
  KEY_NONE,
  KEY_DOWN,
  KEY_UP
};
typedef enum key_state {
  KEY_STATE_IDLE,
  KEY_STATE_DEBOUNCING,
  KEY_STATE_WAIT
};
const int key_debounce_time = 20;
//const int key_repeat_time   = 250;

const bool eeprom_enable                     = true;
const int  eeprom_target_temperature_address = 0;



void setup( )
{
  /* Set up the control pins. */
  pinMode( heater_pin, OUTPUT );
  digitalWrite( heater_pin, LOW );
  pinMode( cooler_pin, OUTPUT );
  digitalWrite( cooler_pin, LOW );
  pinMode( cooler_fan_pin, OUTPUT );
  digitalWrite( cooler_fan_pin, LOW );

  /* Set up the keypad input pins. */
  pinMode( temperature_down_pin, INPUT );
  pinMode( temperature_up_pin, INPUT );

  /* Turn off the on-board LED. */
  pinMode( led_pin, OUTPUT );
  digitalWrite( led_pin, LOW );

  /* Configure the ADC reference voltage to default (5V). */
  analogReference( DEFAULT );

  /* Give the display time to initialize itself. */
  delay( 1000 );
  /* Set up the number of columns and rows of the LCD display. */
  lcd_display.begin( 8, 1 );
  /* Create custom glyphs. */
  lcd_display.createChar( 0, arrow_down );
  lcd_display.createChar( 1, arrow_up );
  lcd_display.createChar( 2, checkmark );

  /* Read the target temperature from EEPROM, if enabled. */
  if( eeprom_enable )
  {
    target_temperature = (float) EEPROM.read( eeprom_target_temperature_address );
  }
  else
  {
    target_temperature = default_temperature;
  }
  if( ( target_temperature < minimum_temperature ) || target_temperature > maximum_temperature )
  {
    target_temperature = default_temperature;
  }


  /* Initialize the serial port. */
  Serial.begin( serial_baud_rate, SERIAL_8N1 );

  set_setpoints( &cooler_setpoint, &heater_setpoint, target_temperature );

  /* Set up the heater PID. */
  heater_PID.SetOutputLimits( 0, heat_period );
  heater_PID.SetMode( AUTOMATIC );

  /* Set up the cooler PID. */
  cooler_PID.SetOutputLimits( 0, cool_period );
  cooler_PID.SetMode( AUTOMATIC );

  /* Initialize the watchdog. */
  wdt_enable( WDTO_1S );
}



void set_setpoints( double *cooler_setpoint, double *heater_setpoint, float setpoint )
{
  *cooler_setpoint = setpoint + setpoint_window;
  *heater_setpoint = setpoint - setpoint_window;
}



/* Print the current temperature and the target temperature on the display.
 *
 * @param [in] temperature Current temperature in degrees Celcius.
 * @param [in] target Target temperature in degrees Celcius.
 * @param [in] heating Flag that indicates that the heater is on.
 * @param [in] cooling Flag that indicates that the cooler is on.
 * @param [in] force_update_display Flag that indicates that the display must
 *        be updated regardless of the update delay.
 */
void display_temperature( float temperature, float target, bool heating, bool cooling, bool force_update_display )
{
  static long display_timestamp = -display_update_interval - 1;

  if( ( (unsigned long)( millis( ) - display_timestamp ) >= display_update_interval )
    || ( force_update_display == true ) )  
  {
    char             display_text[ 9 ];
    char             temperature_text[ 4 ];
    char             target_text[ 3 ];
    custom_character indicator_character;

    /* Limit the temperature range for the target temperature. */
    if( target < minimum_temperature )
    {
      target = minimum_temperature;
    }
    if( target > maximum_temperature )
    {
      target = maximum_temperature;
    }

    /* Determine which temperature glyph to display. */
    if( heating == true )
    {
      indicator_character = ARROW_UP;
    }
    else if( cooling == true )
    {
      indicator_character = ARROW_DOWN;
    }
    else
    {
      indicator_character = CHECKMARK;
    }

    /* Convert the temperatures to integers. */
    int temperature_int = (int)( temperature + 0.5 );
    int target_int      = (int)( target + 0.5 );

    /* Print the temperatures. */
    if( ( temperature < -99.0 ) || ( temperature > 999.0 ) )
    {
      sprintf( temperature_text, "***" );
    }
    else
    {
      sprintf( temperature_text, "%3d", temperature_int );
    }
    if( ( target < minimum_temperature ) || ( target > maximum_temperature ) )
    {
      sprintf( target_text, "**" );
    }
    else
    {
      sprintf( target_text, "%2d", target_int );
    }
    /* Set the cursor to column 0, line 0 and write the current temperature
       and the target temperature. */
    lcd_display.setCursor( 0, 0 );
    sprintf( display_text, "%s (%s)", temperature_text, target_text );
    lcd_display.write( display_text );
    /* Set the cursor to column 7, line 0 and plot the temperature glyph. */
//    lcd_display.setCursor( 7, 0 );
//    lcd_display.write( byte( indicator_character ) );

    display_timestamp = millis( );
  }
}



/* Read the current temperature by samling the MCP9700.
 *
 * @return Current temperature in degrees Celcius.
 */
float read_temperature( )
{
  const float cabinet_offset    = -3.5;
  const float reference_voltage = 4.9;

  int sample  = analogRead( temperature_pin );
#ifdef DEBUG
  mcp9700_voltage = sample;
#endif
  /* Calculate the voltage measured at the MCP9700. */
  float voltage = ( (float)sample / 1024.0 ) * reference_voltage;
  /* Convert the voltage to degrees Celcius according to the datasheet. */
  float temperature = ( voltage - 0.500 ) * 100.0 - cabinet_offset;
  return( temperature );
}



/* Activate the cooler and the cooler fan. */
void activate_cooler( )
{
  if( ! cooling )
  {
#ifndef DEBUG_MEASURE_ONLY
    digitalWrite( cooler_fan_pin, HIGH );
    digitalWrite( cooler_pin, HIGH );
#endif
    cooldown_fan = false;
    cooling = true;
    Serial.println( "Cooler ON" );
  }
}



/* Dectivate the cooler and set a timestamp for the cooler fan to stop. */
void deactivate_cooler( )
{
  if( ! cooldown_fan )
  {
    digitalWrite( cooler_pin, LOW );
    cooldown_fan = true;
    cooler_fan_timestamp = millis( );
  }
  if( cooling )
  {
    cooling = false;
    Serial.println( "Cooler OFF" );
  }
}



/* Turn off the cooler fan some time after the cooler has been turned off. */
void control_cooler_fan( )
{
  /* Only turn off the fan when the cooler has been deactivated and has indicated
     that the cooldown fan should be running for a while. */
  if( cooldown_fan )
  {
    if( (unsigned long)( millis( ) - cooler_fan_timestamp ) >= cooler_fan_delay )
    {
      digitalWrite( cooler_fan_pin, LOW );
      cooldown_fan = false;
    }
  }
}



/* Turn on the heater. */
void activate_heater( )
{
  if( ! heating )
  {
#ifndef DEBUG_MEASURE_ONLY
    digitalWrite( heater_pin, HIGH );
#endif
    heating = true;
    Serial.println( "Heater ON" );
  }
}



/* Turn off the heater. */
void deactivate_heater( )
{
  if( heating )
  {
    digitalWrite( heater_pin, LOW );
    heating = false;
    Serial.println( "Heater OFF" );
  }
}



/* Read a keypress and debounce it.
 *
 * @return Key press (enum key_press).
 */
int read_key_press( )
{
  static long      key_timestamp;
  static key_state key_state    = KEY_STATE_IDLE;
  static key_press current_key  = KEY_NONE;
  key_press        detected_key = KEY_NONE;

  int down_pin_status = digitalRead( temperature_down_pin );
  int up_pin_status   = digitalRead( temperature_up_pin );

  if( key_state == KEY_STATE_IDLE )
  {
    if( down_pin_status == 0 )
    {
      current_key = KEY_DOWN;
      key_state   = KEY_STATE_DEBOUNCING;
    }
    else if( up_pin_status == 0 )
    {
      current_key = KEY_UP;
      key_state   = KEY_STATE_DEBOUNCING;
    }
    
    if( key_state == KEY_STATE_DEBOUNCING )
    {
      key_timestamp = millis( );
    }
  }
  /* Currently debouncing a key. */
  else if( key_state == KEY_STATE_DEBOUNCING )
  {
    /* If the keypress is the same after debouncing, then it's a valid keypress. */
    if( (unsigned long)( millis( ) - key_timestamp ) >= key_debounce_time )
    {
      if( ( current_key == KEY_DOWN ) && ( down_pin_status == 0 ) )
      {
        detected_key = KEY_DOWN;
      }
      else if( ( current_key == KEY_UP ) && ( up_pin_status == 0 ) )
      {
        detected_key = KEY_UP;
      }

      if( detected_key != KEY_NONE )
      {
        key_timestamp = millis( );
        key_state = KEY_STATE_WAIT;
      }
    }
  }
  /* Wait either key_repeat_time milliseconds before allowing another keypress
     or acknowledge that a different reading means the key has been released. */
  else if( key_state == KEY_STATE_WAIT )
  {
    /*
    if( (unsigned long)( millis( ) - key_timestamp ) >= key_repeat_time )
    {
      detected_key = current_key;
    }
    else*/
    if( ( current_key == KEY_DOWN ) && ( down_pin_status == 1 ) )
    {
      current_key = KEY_NONE;
      key_state = KEY_STATE_IDLE;
    }
    else if( ( current_key == KEY_UP ) && ( up_pin_status == 1 ) )
    {
      current_key = KEY_NONE;
      key_state = KEY_STATE_IDLE;
    }
  }

  return( detected_key );
}



/**
 * Adjust the target temperature based on the keypress and store the
 * target temperature in EEPROM.
 *
 * @param keypress Keypress enumeration.
 */
void adjust_target_temperature( int keypress )
{
  key_press key = (key_press)keypress;

  if( key != KEY_NONE )
  {
    if( key == KEY_DOWN )
    {
      target_temperature = target_temperature - 1.0;
      if( target_temperature <= minimum_temperature )
      {
        target_temperature = minimum_temperature;
      }
    }
    else if( key == KEY_UP )
    {
      target_temperature = target_temperature + 1.0;
      if( target_temperature >= maximum_temperature )
      {
        target_temperature = maximum_temperature;
      }
    }

    /* Set the cooler PID and heater PID target temperatures. */
    set_setpoints( &cooler_setpoint, &heater_setpoint, target_temperature );

    if( eeprom_enable )
    {
      EEPROM.write( eeprom_target_temperature_address, (int)target_temperature );
    }
  }
}



/**
 * Write the current temperature, the target temperature, and the cooler and
 * heater activity to the serial port.
 */
void write_status( )
{
  static long status_timestamp = -status_update_interval - 1;

  if( (unsigned long)( millis( ) - status_timestamp ) >= status_update_interval )
  {
#ifdef DEBUG
    Serial.println( (int)( millis( ) / 1000 ) );
#endif
    Serial.print( "Temperature: " );
    Serial.println( current_temperature, 1 );
    Serial.print( "Target: " );
    Serial.println( target_temperature, 1 );
#ifdef DEBUG
//    Serial.print( "Sample: " );
//    Serial.println( mcp9700_voltage );
    Serial.print( "cooler: setpoint = " );
    Serial.print( cooler_setpoint, 1 );
    Serial.print( ", output = " );
    Serial.print( cooler_output, 1 );
    if( cooler_enabled == false ) Serial.print( ", cooler disabled" );
    Serial.println( "" );
    Serial.print( "heater: setpoint = " );
    Serial.print( heater_setpoint, 1 );
    Serial.print( ", output = " );
    Serial.print( heater_output, 1 );
    if( heater_enabled == false ) Serial.print( ", heater disabled" );
    Serial.println( "" );
#endif

    status_timestamp = millis( );
  }
}



int check_shutdown_command( )
{
  char command = -1;

  if( Serial.available( ) )
  {
    char command_read = Serial.read( );
    if( command_read == '0' )
    {
      command = 0;
    }
    else if( command_read == '1' )
    {
      command = 1;
    }
  }

  return command;
}



/* Main loop. */
void loop( )
{
  bool force_update_display = false;
  static long cooler_timestamp = millis( );
  static long heater_timestamp = millis( );
  static bool general_shutdown = false;

  /* Feed the watchdog. */
  wdt_reset( );

  /* Read the current temperature and activate the heater/cooler controls as necessary. */
  current_temperature = read_temperature( );
  /* Control the cooler. */
  cooler_input = current_temperature;
  cooler_PID.Compute( );
  /* Control the heater. */
  heater_input = current_temperature;
  heater_PID.Compute( );

  if( general_shutdown == false )
  {
    /* Disable the cooler PID and the cooler if the temperature is above the lower limit. */
    float current_cooler_output;
    if( current_temperature < target_temperature - temperature_lower_window_low )
    {
      if( cooler_enabled == true )
      {
        cooler_PID.SetMode( MANUAL );
        deactivate_cooler( );
        cooler_enabled = false;
      }
      current_cooler_output = 0.0;
    }
    else if( current_temperature > target_temperature - temperature_lower_window )
    {
      if( cooler_enabled == false )
      {
        cooler_PID.SetMode( AUTOMATIC );
        cooler_enabled = true;
      }
      current_cooler_output = cooler_output;
    }
    /* Set the cooler time segment. */
    if( (unsigned long)( millis( ) - cooler_timestamp ) > cool_period )
    {
      cooler_timestamp += cool_period;
    }
    if( cooler_enabled == true )
    {
      if( ( (unsigned long)( millis( ) - cooler_timestamp ) <= (unsigned long)current_cooler_output )
        && ( current_cooler_output > lower_cool_period ) )
      {
        activate_cooler( );
      }
      else if( current_cooler_output < upper_cool_period )
      {
        deactivate_cooler( );
      }
    }
  
    /* Disable the heater PID and the heater if the temperature is above the upper limit. */
    float current_heater_output;
    if( current_temperature > target_temperature + temperature_upper_window_high )
    {
      if( heater_enabled == true )
      {
        heater_PID.SetMode( MANUAL );
        deactivate_heater( );
        heater_enabled = false;
        Serial.println( "Heater disabled" );
      }
      current_heater_output = 0.0;
    }
    else if( current_temperature < target_temperature + temperature_upper_window )
    {
      if( heater_enabled == false )
      {
        heater_PID.SetMode( AUTOMATIC );
        heater_enabled = true;
        Serial.println( "Heater enabled" );
      }
      current_heater_output = heater_output;
    }
 
    /* Set the heater time segment. */
    if( (unsigned long)( millis( ) - heater_timestamp ) > heat_period )
    {
      heater_timestamp = millis( );
    }
    if( heater_enabled == true )
    {
      if( ( (unsigned long)( millis( ) - heater_timestamp ) <= (unsigned long)current_heater_output )
        && ( current_heater_output > lower_heat_period ) )
      {
        activate_heater( );
      }
      else if( current_heater_output < upper_heat_period )
      {
        deactivate_heater( );
      }
    }
  }

  /* Determine if we've received a shutdown (0) or wakeup (1) command. */
  int shutdown_check = check_shutdown_command( );
  if( shutdown_check == 0 )
  {
    general_shutdown = true;
    cooler_PID.SetMode( MANUAL );
    deactivate_cooler( );
    cooler_output  = 0.0;
    cooler_enabled = false;
    heater_PID.SetMode( MANUAL );
    deactivate_heater( );
    heater_output  = 0.0;
    heater_enabled = false;
    Serial.println( "Climate box shutdown" );
  }
  else if( shutdown_check == 1 )
  {
    general_shutdown = false;
    Serial.println( "Climate box turned on" );
  }

  /* Turn off the cooler fan if the cooler is off and some time has passed. */
  control_cooler_fan( );

  /* Get the user's keypress input, if any, and adjust the target temperature
     accordingly. */
  key_press key = (key_press)read_key_press( );
  adjust_target_temperature( key );
  if( key != KEY_NONE )
  {
    force_update_display = true;
  }

  /* Update the display. */
  display_temperature( current_temperature, target_temperature, heating, cooling,
                       force_update_display );

  /* Write status information over the serial port. */
  write_status( );
}

